"""
A quest is a distributed scraping journey :)
A sciop instance contains a list of dataset parts to scrape,
a crawler will get a series of unclaimed parts,
crawl them, and upload the resulting torrents!
"""

from __future__ import annotations

import json
import warnings
from pathlib import Path
from typing import Any, ClassVar, Literal, Self, cast

from pydantic import BaseModel, Field, model_validator
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from sciop_cli.api import create_upload
from sciop_cli.data import get_default_trackers
from sciop_cli.models.sciop import Upload
from sciop_cli.torrent import create_torrent, find_optimal_piece_size
from scrapy import Spider
from scrapy.crawler import CrawlerProcess

from sciop_scraping.const import METHOD_STR
from sciop_scraping.exceptions import QuestInvalidUploadError


class Quest(BaseModel):
    """
    A scrape journey to go on with your friends
    """

    name: ClassVar[str]
    """
    Name that will be used when calling from the CLI.
    Each quest needs an (instance) unique name
    """
    dataset_slug: ClassVar[str]
    """
    Dataset slug on the sciop instance that this quest belongs to.
    
    Usually the same as `name`, but allowed to differ e.g.
    in the case of multiple quests for the same dataset.
    """
    part_slugs: list[str] | None = None
    """
    Part slugs that this subquest corresponds to.
    
    Typically, but not always, equivalent to `subquest`
    """
    spider: ClassVar[type[Spider] | None] = None
    """
    Scrapy spider to use, if any
    """

    output: Path
    """
    Output directory to write to.
    
    The output directory stores data for *all* quests and subquests.
    
    Quests should ensure that their scraper outputs data in a structure like..
    
    ```
    output
    |- quest_status.json
    |- quest_1
    |  |- subquest_1
    |  |  |- (data)
    |  |  ...
    |  |- subquest_1
    |  ...
    |- quest_2
    ...
    ```
    """
    subquest: str
    """
    The currently active subquest.
    """

    def run(
        self, scrape_kwargs: dict | None = None, process_kwargs: dict | None = None
    ) -> QuestStatus:
        if scrape_kwargs is None:
            scrape_kwargs = {}
        if process_kwargs is None:
            process_kwargs = {}

        status = self.init_status()
        try:
            if status.status == "scraping":
                status, scrape_kwargs = self.before_scrape(status, **scrape_kwargs)
                status = self.scrape(status, scrape_kwargs, process_kwargs)
                status = self.after_scrape(status)
                status.status = "validating"
            if status.status == "validating":
                status = self.before_validate(status)
                status = self.validate_scrape(status)
                if not status.validation_errors:
                    status.status = "packing"
                else:
                    status.result = "validation_error"
                status = self.after_validate(status)
            if status.status == "packing":
                status = self.before_pack(status)
                status = self.pack(status)
                status = self.after_pack(status)
                status.status = "complete"
            if status.result is None:
                # status would otherwise be set to validation_error in validate_scrape
                status.result = "success"
        finally:
            self.update_log(status)

        return status

    def init_status(self) -> QuestStatus:
        """
        Load any intermediate result from the quest log,
        or create a new one if none found.

        If we have previously attempted a subquest and ended with a validation error,
        retry from the scraping stage.
        """
        status = None
        if self.log_path.exists():
            log = QuestLog.from_json(self.log_path)
            status = log.get_subquest(self.name, self.subquest)

        if status is None:
            status = QuestStatus(
                quest=self.name,
                subquest=self.subquest,
                path=self.subquest_path,
                dataset_slug=self.dataset_slug,
                part_slugs=[self.subquest] if not self.part_slugs else self.part_slugs,
            )
        else:
            # if we have previously completed with a validation error and are running again,
            # return to scraping state.
            if status.result == "validation_error":
                status.status = "scraping"

        status.n_attempts += 1

        return status

    @property
    def log_path(self) -> Path:
        return self.output / "quest-log.json"

    @property
    def subquest_path(self) -> Path:
        return self.output / self.name / self.subquest

    def update_log(self, res: QuestStatus) -> None:
        """
        Update the log, replacing the
        :param res:
        :return:
        """
        log = QuestLog.from_json(self.log_path)
        log = log.update(res)
        log.to_json(self.log_path)

    def scrape(self, res: QuestStatus, scrape_kwargs: dict, process_kwargs: dict) -> QuestStatus:
        """
        Run the spider - default is to use the scrapy spider
        """
        return self._scrape_with_spider(res, scrape_kwargs, process_kwargs)

    def validate_scrape(self, res: QuestStatus) -> QuestStatus:
        """
        After scraping, subclasses may override to perform validation.

        Default is no-op
        """
        return res

    def pack(self, res: QuestStatus) -> QuestStatus:
        """
        After validating, subclasses may override to perform packing.

        Default is no-op
        """
        return res

    def before_scrape(self, res: QuestStatus, **kwargs: any) -> tuple[QuestStatus, dict]:
        """Hook method called before scraping"""
        return res, kwargs

    def after_scrape(self, res: QuestStatus) -> QuestStatus:
        """Hook method called after scraping"""
        return res

    def before_validate(self, res: QuestStatus) -> QuestStatus:
        """Hook method called before validation"""
        return res

    def after_validate(self, res: QuestStatus) -> QuestStatus:
        """Hook method called after validation"""
        return res

    def before_pack(self, res: QuestStatus) -> QuestStatus:
        """Hook method called before packing"""
        return res

    def after_pack(self, res: QuestStatus) -> QuestStatus:
        """Hook method called after packing"""
        return res

    def _scrape_with_spider(
        self, res: QuestStatus, scrape_kwargs: dict, process_kwargs: dict
    ) -> QuestStatus:
        if self.spider is None:
            raise RuntimeError("No spider has been declared, but attempted to scrape with spider")

        process = CrawlerProcess(**process_kwargs)
        process.crawl(self.spider, **scrape_kwargs)
        process.start(install_signal_handlers=False)
        return res


class QuestStatus(BaseModel):
    quest: str
    subquest: str
    path: Path
    """Directory where the data was scraped to"""
    dataset_slug: str
    """
    Corresponding dataset in the sciop instance
    Usually equivalent to `quest`, but allowed to differ 
    e.g. for multiple quests for the same dataset.
    """
    part_slugs: list[str] | None
    """
    Part slugs that this subquest corresponds to.
    Allowed to be separable from subquest so that a subquest might refer to multiple dataset parts,
    or their names may differ.
    """
    scrape_errors: list[ScrapeError] = Field(default_factory=list)
    validation_errors: list[ValidationError] = Field(default_factory=list)
    status: Literal["scraping", "validating", "packing", "complete", "uploaded", "disabled"] = (
        "scraping"
    )
    n_attempts: int = 0
    result: None | Literal["success", "validation_error", "impossible"] = None

    _TABLE_COLS: ClassVar[tuple[str]] = ("subquest", "status", "result", "errors", "n_attempts")

    @model_validator(mode="before")
    @classmethod
    def fill_dataset_and_part_slugs(cls, data: dict | QuestStatus) -> dict | QuestStatus:
        """
        If dataset_slug or part_slugs are missing (e.g. from old version of quest status format,
        fill them.
        """
        if isinstance(data, dict):
            if "dataset_slug" not in data:
                data["dataset_slug"] = data["quest"]
            if "part_slugs" not in data:
                data["part_slugs"] = [data["subquest"]]
        elif isinstance(data, BaseModel):
            data = cast(QuestStatus, data)
            if not data.dataset_slug:
                data.dataset_slug = data.quest
            if not data.part_slugs:
                data.part_slugs = [data.subquest]
        return data

    @property
    def total_errors(self) -> int:
        return len(self.scrape_errors) + len(self.validation_errors)

    def to_comment(self) -> str:
        """Format quest status as a string to be used in a torrent comment"""
        dumped = self.model_dump(
            exclude_none=True, exclude={"path", "scrape_errors", "validation_errors"}
        )
        dumped["scrape_errors"] = len(self.scrape_errors)
        dumped["validation_errors"] = len(self.validation_errors)
        comment_parts = [f"Scraped with {METHOD_STR}"]
        comment_parts.extend([f"{key}: {value}" for key, value in dumped.items()])
        return "\n".join(comment_parts)

    def table_row(self) -> list[Text]:
        style_map = {
            "bright_green": {"success", "complete"},
            "bright_red": {"validation_error", "impossible"},
            "bright_blue": {"uploaded"},
        }
        styles = {value: style for style, values in style_map.items() for value in values}
        row = []
        for key in self._TABLE_COLS:
            if key == "errors":
                row.append(
                    Text(
                        str(self.total_errors),
                        style="bright_red" if self.total_errors > 0 else None,
                    )
                )
            else:
                value = str(getattr(self, key))
                row.append(Text(value, style=styles.get(value)))
        return row


class QuestLog(BaseModel):
    subquests: list[QuestStatus] = Field(default_factory=list)

    def get_subquest(self, quest: str, subquest: str) -> QuestStatus | None:
        matches = [q for q in self.subquests if q.quest == quest and q.subquest == subquest]
        if len(matches) > 1:
            raise KeyError(f"More than one match found for quest {quest}, subquest {subquest}")
        elif len(matches) == 1:
            return matches[0]
        else:
            return None

    def update(self, res: QuestStatus) -> Self:
        """
        Add new entry to log, replacing any previous matches
        """
        items = [q for q in self.subquests if q.quest != res.quest or q.subquest != res.subquest]
        existing = [
            q for q in self.subquests if q.quest == res.quest and q.subquest == res.subquest
        ]
        if existing:
            res = self.merge_errors(res, existing[0])
        items.append(res)
        self.subquests = items
        return self

    def merge_errors(self, res: QuestStatus, existing: QuestStatus) -> QuestStatus:
        """
        Merge errors from the previous status, updating the n_errors in each.
        """
        old_scrape_map = {e.url: e for e in existing.scrape_errors}
        old_validation_map = {e.path: e for e in existing.validation_errors}

        for e in res.scrape_errors:
            if e.url in old_scrape_map:
                e.n_failures = old_scrape_map[e.url].n_failures + 1
        for e in res.validation_errors:
            if e.path in old_validation_map:
                e.n_failures = old_validation_map[e.path].n_failures + 1
        return res

    @classmethod
    def from_json(cls, path: Path) -> QuestLog:
        if not path.exists():
            return QuestLog()

        try:
            with open(path) as f:
                log = json.load(f)
        except json.decoder.JSONDecodeError:
            warnings.warn(
                f"Quest log could not be read from {str(path)}, ignoring, will overwrite",
                stacklevel=2,
            )
            log = []
        return QuestLog(subquests=log)

    def to_json(self, path: Path) -> None:
        items = self.model_dump(by_alias=True)["subquests"]
        with open(path, "w") as f:
            json.dump(items, f, indent=2, default=str)

    def to_tables(self) -> dict[str, Table]:
        tables = {}
        for subquest in self.subquests:
            if subquest.quest not in tables:
                tables[subquest.quest] = Table(*QuestStatus._TABLE_COLS, title=subquest.quest)
            tables[subquest.quest].add_row(*subquest.table_row())
        return tables

    def __rich__(self) -> Panel:
        group = Group(*self.to_tables().values())
        return Panel(group, title="sciop quest log")


class ScrapeError(BaseModel):
    url: str
    type_: str = Field(..., alias="type")
    msg: str
    n_failures: int = 1


class ValidationError(BaseModel):
    path: Path
    type_: str = Field(..., alias="type")
    msg: str
    n_failures: int = 1


def _make_subquest_torrent(
    status: QuestStatus,
    torrent_path: Path,
    progress: bool = False,
    torrent_kwargs: dict | None = None,
) -> Path:
    """
    Make a torrent for a subquest.
    This ignores completeness/validation errors and just makes the torrent,
    it should typically be called from upload_subquest instead
    """
    default_torrent_kwargs: dict[str, Any] = {
        "version": "hybrid",
        "bencode": True,
        "pbar": progress,
        "comment": status.to_comment(),
    }
    if torrent_kwargs is None:
        torrent_kwargs = default_torrent_kwargs
    else:
        torrent_kwargs = {**default_torrent_kwargs, **torrent_kwargs}

    if "piece_size" not in torrent_kwargs:
        torrent_kwargs["piece_size"] = find_optimal_piece_size(
            status.path, version=torrent_kwargs["version"]
        )
    if "trackers" not in torrent_kwargs:
        torrent_kwargs["trackers"] = get_default_trackers()

    bencoded = create_torrent(path=status.path, **torrent_kwargs)
    with open(torrent_path, "wb") as f:
        f.write(bencoded)
    return torrent_path


def upload_subquest(
    status: QuestStatus,
    torrent_dir: Path,
    torrent_path: Path | None = None,
    force: bool = False,
    progress: bool = False,
    torrent_kwargs: dict = None,
) -> tuple[Upload, Path]:
    """
    Upload a subquest that has been completed!

    Create a torrent, then upload it to the configured sciop instance using a local sciop-cli config

    args:
        status (QuestStatus): The status object containing information about the given subquest
        torrent_dir (Path): Directory where the torrent will be created
        force (bool, optional): Upload even if quest was incomplete or not validated,
            and recreate torrent even if it's absent
        progress (bool): Display progress bar while creating torrent
        torrent_kwargs (dict, optional): Keyword arguments to pass to sciop-cli create_torrent
    """
    if (status.status != "complete" or status.result != "success") and not force:
        raise QuestInvalidUploadError(
            f"Quest could not be uploaded when it is incomplete or had validation errors: {status}"
        )

    if torrent_path is None:
        torrent_path = torrent_dir / (status.subquest + ".torrent")
    if torrent_path.exists() and not force:
        warnings.warn(
            f"Expected output torrent already exists and force was False, not recreating torrent.\n"
            f"path: {torrent_path}",
            stacklevel=2,
        )
    else:
        _make_subquest_torrent(
            status=status,
            torrent_path=torrent_path,
            progress=progress,
            torrent_kwargs=torrent_kwargs,
        )

    upload = create_upload(
        dataset=status.dataset_slug,
        dataset_parts=status.part_slugs,
        torrent_path=torrent_path,
        method=METHOD_STR,
    )

    # upload subquest status and save to quest log
    status.status = "uploaded"
    quest_log = QuestLog.from_json(torrent_dir.parent / "quest-log.json")
    quest_log.update(status)
    quest_log.to_json(path=torrent_dir.parent / "quest-log.json")

    return upload, torrent_path



import json
from pathlib import Path

import click
from scrapy.crawler import CrawlerProcess

from sciop_scraping.quests.us_oig.spider import OIGSpider


@click.command("us-oig")
@click.option(
    "-o",
    "--output",
    help="Output directory to save files in. "
    "If None, $PWD/data/us-oig. "
    "Data will be saved in a chronicling-america subdirectory, "
    "and the crawl state will be saved in crawl_state.",
    default=None,
    type=click.Path(),
)
def us_oig(output: str | None = None) -> None:
    output = Path.cwd() / "data" / "us-oig" if not output else Path(output)
    output.mkdir(exist_ok=True, parents=True)
    meta_path = output / "metadata.jsonl"
    if meta_path.exists():
        metadata = []
        with open(meta_path) as f:
            for line in f:
                metadata.append(json.loads(line))
    else:
        metadata = []

    process = CrawlerProcess(settings={"FEEDS": {meta_path: {"format": "jsonlines"}}})
    process.crawl(OIGSpider, output=output, metadata=metadata)
    process.start()



"""
Mixins for filtering urls, pages, files, etc.
"""

import re
from collections.abc import Iterable
from typing import TypeAlias

PatternArgType: TypeAlias = str | re.Pattern | list[str] | list[re.Pattern] | None


class URLFilterMixin:
    """
    Filter urls according to include/exclude patterns.

    Like scrapy's Link extractors, except works on sets of urls as strings,
    rather than scrapy's `Response` objects and HTML responses only.

    Written using `classmethods` for use as a mixin,
    to use by composition as a an independent object,
    use :class:`URLFilter`

    `include` and `exclude` are regex patterns, or lists of regex patterns,
    and urls are filtered such that
    - If neither include or exclude are provided, no filtration is done
    - If only `include` is provided, only urls that match one of the patterns are included.
    - If only `exclude` is provided, all urls *except* those that match one of the patterns
      are excluded.
    - If both are provided, all urls that match one of the `include` patterns
      and do not match any of the `exclude` patterns are included.

    Both sets of filters are `OR`'d - urls match against *any* of them, rather than *all* of them.
    Both sets of filters are matched with a *full match* -
    the entire string must match the pattern for a hit.
    For partial matching, use wildcards and string start/end anchors.
    """

    @classmethod
    def matches_pattern(cls, value: str, pattern: PatternArgType, flags: int | None = None) -> bool:
        """Check if a string matches a pattern or set of patterns"""
        if pattern is None:
            return True
        if not isinstance(pattern, list):
            pattern = [pattern]
        pattern = [p if isinstance(p, re.Pattern) else re.compile(p, flags=flags) for p in pattern]
        return any([p.fullmatch(value) for p in pattern])

    @classmethod
    def url_valid(
        cls,
        url: str,
        include: PatternArgType = None,
        exclude: PatternArgType = None,
        flags: int | None = None,
    ) -> bool:
        """Check if a string is valid given a set of include and exclude patterns"""
        if not exclude:
            # if exclude is None, the string always trivially matches it,
            # which is the opposite of what we want for excludes.
            return cls.matches_pattern(url, include, flags)
        else:
            return cls.matches_pattern(url, include, flags) and not cls.matches_pattern(
                url, exclude, flags
            )

    @classmethod
    def filter_urls(
        cls,
        urls: list[str],
        include: PatternArgType = None,
        exclude: PatternArgType = None,
        flags: int | None = None,
    ) -> list[str]:
        """Filter a list of urls according to include and exclude patterns"""
        return [url for url in urls if cls.url_valid(url, include, exclude, flags)]

    @classmethod
    def iter_filtered_urls(
        cls,
        urls: Iterable[str],
        include: PatternArgType = None,
        exclude: PatternArgType = None,
        flags: int | None = None,
    ) -> Iterable[str]:
        """Iterate over some iterable of urls, yielding only valid urls"""
        for url in urls:
            if cls.url_valid(url, include, exclude, flags):
                yield url


class URLFilter:
    """
    Independent version of :class:`URLFilterMixin`,
    that accepts include/exclude patterns as params,
    and uses the URLFilterMixin class when parsing urls.
    """

    def __init__(
        self,
        include: PatternArgType = None,
        exclude: PatternArgType = None,
        flags: int | None = None,
    ):
        self.include: list[re.Pattern] | None = _pattern_input_to_pattern_list(include, flags)
        self.exclude: list[re.Pattern] | None = _pattern_input_to_pattern_list(exclude, flags)
        self.flags = flags

    def url_valid(self, url: str) -> bool:
        return URLFilterMixin.url_valid(url, self.include, self.exclude, self.flags)

    def filter_urls(self, urls: list[str]) -> list[str]:
        return URLFilterMixin.filter_urls(urls, self.include, self.exclude, self.flags)

    def iter_filtered_urls(self, urls: Iterable[str]) -> Iterable[str]:
        yield from URLFilterMixin.iter_filtered_urls(urls, self.include, self.exclude, self.flags)


def _pattern_input_to_pattern_list(
    pattern: PatternArgType, flags: int | None = None
) -> list[re.Pattern] | None:
    if pattern is None:
        return pattern
    if not isinstance(pattern, list):
        pattern = [pattern]
    pattern = [p if isinstance(p, re.Pattern) else re.compile(p, flags=flags) for p in pattern]
    return pattern



"""
Helper mixin for scraping files (and maybe pages) from archive.org's wayback machine
"""

from collections.abc import Callable, Generator
from datetime import datetime
from typing import Any, Self, TypedDict
from urllib.parse import quote

import scrapy
from scrapy.http import JsonRequest, JsonResponse
from scrapy.responsetypes import Response

from sciop_scraping.mixins.filter import URLFilter


class WaybackDatetime(datetime):
    """
    Wayback machine uses flat timestamps that are like
    YYYYMMDDHHMMSS with no delimiters.

    Simple convenience methods for parsing/rendering
    """

    WAYBACK_STRF = "%Y%m%d%H%M%S"

    @staticmethod
    def format_str(timestamp: datetime) -> str:
        return timestamp.strftime(WaybackDatetime.WAYBACK_STRF)

    @classmethod
    def from_str(cls, timestamp: str) -> Self:
        return cls.strptime(timestamp, cls.WAYBACK_STRF)

    def to_str(self) -> str:
        return self.strftime(self.WAYBACK_STRF)


class TimeMapDict(TypedDict, total=False):
    """
    Response type when requesting "time maps",
    or the set of urls beneath a given prefix
    """

    original: str
    """url of the original page"""
    mimetype: str
    timestamp: str
    endtimestamp: str
    groupcount: str
    uniqcount: str


class WaybackMixin:
    TIMEMAP_TEMPLATE = "https://web.archive.org/web/timemap/json?url={url}&matchType=prefix&collapse=urlkey&output=json&fl=original%2Cmimetype%2Ctimestamp%2Cendtimestamp%2Cgroupcount%2Cuniqcount&filter=!statuscode%3A%5B45%5D..&limit=10000"

    def wb_get_prefix(
        self,
        url: str,
        callback: Callable[[list[str]], Any],
        **kwargs: Any,
    ) -> Generator[JsonRequest, None, None]:
        """
        Get the urls for the latest snapshots of files beneath some url prefix,
        returning the list of urls to a given callback function.

        To call a parser callback for *requested* copies of each of the urls,
        use :meth:`.wb_expand_prefix`
        """
        timemap_url = self.TIMEMAP_TEMPLATE.format(url=quote(url))
        yield JsonRequest(
            timemap_url,
            callback=self._wb_extract_prefix_urls,
            cb_kwargs={"callback": callback},
            **kwargs,
        )

    def _wb_extract_prefix_urls(
        self, response: JsonResponse, callback: Callable[[list[str]], Generator[Any, None, None]]
    ) -> None:

        res_json = response.json()
        headers = res_json[0]
        items: list[TimeMapDict] = [
            {key: val for key, val in zip(headers, item)} for item in res_json[1:]
        ]
        urls = [
            self.wb_url_from_timestamp(item["original"], item["endtimestamp"]) for item in items
        ]
        yield from callback(urls)

    def wb_expand_prefix(
        self,
        url: str,
        callback: Callable[[Response], Any],
        url_filter: URLFilter | None = None,
        request_kwargs: dict | None = None,
    ) -> Generator[JsonRequest, None, None]:
        """
        Given some url prefix, get the latest version of all urls that match some url filter,
        and then process with some passed callback method.

        First gather the urls by making a request to an undocumented wayback api,
        then yield requests for each matching url using :meth:`.wb_parse_prefix_json`,
        then the results of those urls are passed to the provided callback method.

        The :class:`.TimeMap` dict that contains metadata for the item,
        including the timestamp of the snapshot, will be in the request's
        `meta` attr with a `"timemap"` key.
        """
        if request_kwargs is None:
            request_kwargs = {}
        timemap_url = self.TIMEMAP_TEMPLATE.format(url=quote(url))

        yield scrapy.JsonRequest(
            timemap_url,
            callback=self.wb_parse_prefix_json,
            dont_filter=True,
            cb_kwargs={
                "callback": callback,
                "url_filter": url_filter,
                "request_kwargs": request_kwargs,
            },
            **request_kwargs,
        )

    def wb_parse_prefix_json(
        self,
        response: JsonResponse,
        callback: Callable[[Response], Any],
        url_filter: URLFilter | None = None,
        request_kwargs: dict | None = None,
    ) -> Generator[scrapy.Request, None, None]:
        """
        Parse the response from a wb_expand_prefix request for a "timemap" of urls beneath a prefix,
        requesting any that match a filter with the provided callback.

        The format of the response is a table in the form of a list of lists,
        where the first item in the list are the keys for each of the subsequent lists.

        so like

        ```
        [
            ["letter", "number", "whatever"],
            ["a", 1, "something"]
        ]
        ```

        corresponds to

        ```
        {
            "letter": "a",
            "number": 1,
            "whatever": "something",
        }
        ```

        and the fields from this response correspond to :class:`.TimeMapDict`
        """
        if request_kwargs is None:
            request_kwargs = {}

        res_json = response.json()
        headers = res_json[0]
        items: list[TimeMapDict] = [
            {key: val for key, val in zip(headers, item)} for item in res_json[1:]
        ]
        for item in items:
            if url_filter is not None and not url_filter.url_valid(item["original"]):
                continue
            yield response.follow(
                self.wb_url_from_timestamp(item["original"], item["endtimestamp"]),
                callback=callback,
                dont_filter=True,
                meta={"timemap": item},
                **request_kwargs,
            )

    def wb_follow(
        self, response: Response, callback: Callable, **kwargs: Any
    ) -> Generator[scrapy.Request, None, None]:
        """
        Follow some response and get the latest wayback machine copy of that url,
        calling some callback with the result
        """
        yield response.follow(
            f"https://web.archive.org/{quote(response.url)}", callback=callback, **kwargs
        )

    @staticmethod
    def wb_url_from_timestamp(url: str, timestamp: datetime | str) -> str:
        """
        Construct a wayback machine url from the target page url and a timestamp.

        Just a string formatting method, use :meth:`.get_snapshots` to find valid timestamps.

        Args:
            url: Target page url
            timestamp: Timestamp of a snapshot either as a datetime object or a
                :class:`.WaybackDatetime` YYYYMMDDHHMMSS string

        """
        if isinstance(timestamp, datetime):
            timestamp = WaybackDatetime.format_str(timestamp)

        return f"https://web.archive.org/web/{timestamp}/{url}"



"""
Common options for cli commands
"""

from pathlib import Path

import click


def spider_options(f: click.Command) -> click.Command:
    """Common options that all spider cli commands should support"""
    f = click.option("--retries", type=int, help="Number of retries")(f)
    f = click.option("--timeout", type=int, help="Timeout for spider in seconds")(f)
    return f


def log_path(f: click.Command) -> click.Command:
    default_logfile = Path.cwd() / "data" / "quest-log.json"
    f = click.option(
        "-l",
        "--quest-log",
        type=click.Path(exists=True, file_okay=True, dir_okay=False),
        help="Path to quest log.",
        default=default_logfile,
        show_default=True,
    )(f)
    return f



import hashlib
import json
import re
from pathlib import Path
from typing import Any, Literal

from tqdm import tqdm

from sciop_scraping.quests import Quest, QuestStatus, ValidationError
from sciop_scraping.quests.smithsonian.spider import SmithsonianSpider


class SmithsonianQuest(Quest):
    name = "smithsonian"
    spider = SmithsonianSpider

    dataset_slug: str
    format: Literal["jpg", "tif"]

    def before_scrape(self, res: QuestStatus, **kwargs: Any) -> tuple[QuestStatus, dict]:
        kwargs["format"] = self.format
        kwargs["dataset"] = self.archive_name
        kwargs["output"] = self.output
        return res, kwargs

    def validate_scrape(self, res: QuestStatus) -> QuestStatus:
        res.validation_errors = []
        with open(self.meta_path) as f:
            lines = f.readlines()
        if len(lines) == 0:
            res.validation_errors.append(
                ValidationError(path=self.data_path, type="no_files", msg="No files found!")
            )

        meta = [json.loads(line) for line in lines]
        for item in tqdm(meta, desc="Validating item presence"):
            key = re.sub(r"^media/", "", item["key"])
            path = self.output / "smithsonian" / f"{self.archive_name}-{self.format}" / key
            if not path.exists():
                res.validation_errors.append(
                    ValidationError(path=path, type="missing", msg="Expected file not found!")
                )

        if res.validation_errors:
            return res

        for item in tqdm(meta, desc="Validating item hashes"):
            key = re.sub(r"^media/", "", item["key"])
            path = self.output / "smithsonian" / f"{self.archive_name}-{self.format}" / key
            with open(path, "rb") as f:
                file_hash = hashlib.file_digest(f, "md5").hexdigest()
            if file_hash != item["etag"]:
                res.validation_errors.append(
                    ValidationError(
                        path=path, type="invalid", msg="File did not match etag and was removed!"
                    )
                )
                path.unlink(missing_ok=True)

        return res

    def pack(self, res: QuestStatus) -> QuestStatus:
        """Sort the item metadata"""
        items = {}
        with open(self.meta_path) as f:
            for line in f:
                parsed = json.loads(line)
                if parsed:
                    items[parsed["key"]] = parsed
        items = dict(sorted(items.items()))
        with open(self.meta_path, "w") as f:
            for item in items.values():
                f.write(json.dumps(item) + "\n")
        return res

    @property
    def archive_name(self) -> str:
        return re.sub(r"^si-", "", self.dataset_slug)

    @property
    def data_path(self) -> Path:
        return (
            self.output / "smithsonian" / f"{self.archive_name}-{self.format}" / self.archive_name
        )

    @property
    def subquest_path(self) -> Path:
        return self.data_path

    @property
    def meta_path(self) -> Path:
        return self.data_path / "metadata.jsonl"



from sciop_scraping.quests.base import Quest, QuestLog, QuestStatus, ScrapeError, ValidationError
from sciop_scraping.quests.chronicling.quest import ChroniclingAmericaQuest

__all__ = [
    "ChroniclingAmericaQuest",
    "Quest",
    "QuestLog",
    "QuestStatus",
    "ScrapeError",
    "ValidationError",
]



import click

from sciop_scraping.cli.api import sciop_api
from sciop_scraping.cli.state import state_cli
from sciop_scraping.quests.chronicling.cli import chronicling_america
from sciop_scraping.quests.poes.cli import noaa_poes
from sciop_scraping.quests.smithsonian.cli import smithsonian
from sciop_scraping.quests.us_oig.cli import us_oig


@click.group("sciop-scrape")
def cli() -> None:
    """Distributed scraping with sciop :)"""
    pass


# global groups
cli.add_command(state_cli)
cli.add_command(sciop_api)

# quest groups
cli.add_command(chronicling_america)
cli.add_command(noaa_poes)
cli.add_command(us_oig)
cli.add_command(smithsonian)



from pathlib import Path

import click
from rich.console import Console

from sciop_scraping.cli.common import log_path
from sciop_scraping.quests.base import QuestLog


@click.group("state", invoke_without_command=True)
@click.pass_context
@log_path
def state_cli(
    ctx: click.Context,
    quest_log: str | Path,
) -> None:
    """
    Display or modify the state in a quest log.

    To display, call with no arguments or a --quest-log,
    otherwise, see --help of subcommands
    """
    quest_log = Path(quest_log)
    ctx.ensure_object(dict)
    ctx.obj["quest_log"] = quest_log
    if ctx.invoked_subcommand is None:
        log = QuestLog.from_json(quest_log)
        console = Console()
        console.print(log)



"""
Spider to crawl reports from the office of the inspector general
"""

import multiprocessing as mp
import random
from collections.abc import Generator
from pathlib import Path

import scrapy
from scrapy.http import HtmlResponse, Request, Response
from scrapy.settings import BaseSettings

from sciop_scraping.const import USER_AGENTS
from sciop_scraping.mixins.wayback import WaybackMixin

BASE_URL = "https://www.stateoig.gov/reports"
PAGE_TEMPLATE = BASE_URL + "?page={n}"
PDF_ROOT = "https://www.stateoig.gov/uploads/report/"


class ReportMeta(scrapy.Item):
    report_id = scrapy.Field()
    title = scrapy.Field()
    timestamp = scrapy.Field()
    url = scrapy.Field()
    description = scrapy.Field()
    terms = scrapy.Field()
    files = scrapy.Field()
    file_urls = scrapy.Field()


class OIGSpider(scrapy.Spider, WaybackMixin):
    """
    Outputs
    - a `metadata.jsonl` file containing the title, description, date and
      terms for each of the reports as well as a map to the files
    - a `files/` directory containing the pdfs
    - a `archive/` directory that contains all PDFs from archive.org,
      potentially including duplicates of the files from the main site.
    """

    name = "us-oig"
    custom_settings = {
        "RETRY_ENABLED": True,
        "RETRY_TIMES": 10,
        "RETRY_HTTP_CODES": [500, 502, 503, 504, 522, 524, 408],
        "CONCURRENT_REQUESTS": int(round(mp.cpu_count() * 1.5)),
        "CONCURRENT_REQUESTS_PER_DOMAIN": int(round(mp.cpu_count() * 1.5)),
    }

    def __init__(self, output: Path | None = None, metadata: list[dict] = None):
        super().__init__()
        if output is None:
            output = Path().cwd() / "data" / "us-oig"
            output.mkdir(exist_ok=True)
        self.output = output
        (self.output / "archive").mkdir(exist_ok=True)
        (self.output / "files").mkdir(exist_ok=True)
        self.metadata = {item["report_id"]: item for item in metadata}

    @classmethod
    def update_settings(cls, settings: BaseSettings) -> None:
        settings.set("USER_AGENT", random.choice(USER_AGENTS))
        settings.set("ROBOTSTXT_OBEY", False, priority="spider")

    @property
    def metadata_path(self) -> Path:
        return self.output / "metadata.json"

    async def start(self) -> Generator[Request, None, None]:
        self.logger.info(self.settings["USER_AGENT"])
        yield scrapy.Request(PAGE_TEMPLATE.format(n=0), callback=self.parse_first)
        # yield next(self.wb_get_prefix(PDF_ROOT, self.parse_archive_urls))

    def parse_first(self, response: HtmlResponse) -> Generator[Request, None, None]:
        """
        Parse the first page first to get the max number of pages,
        then parse normally
        """
        self.parse_page(response)
        buttons = response.css("a.usa-pagination__button::text").getall()
        page_ns = []
        for button in buttons:
            try:
                page_ns.append(int(button.strip()))
            except ValueError:
                continue
        max_page = max(page_ns)
        for page_n in range(1, max_page):
            yield scrapy.Request(PAGE_TEMPLATE.format(n=page_n), callback=self.parse_page)

    def parse_page(self, response: HtmlResponse) -> Generator[Request, None, None]:
        links = response.css("div.views-row a::attr(href)").getall()
        for link in links:
            report_n = link.split("/")[-1]
            if report_n in self.metadata:
                continue
            yield response.follow(link, callback=self.parse_report)

    def parse_report(self, response: HtmlResponse) -> Generator[Request | ReportMeta, None, None]:
        report_id = response.url.split("/")[-1]
        title = response.css("h1.page-title span::text").get()
        timestamp = response.css("time::attr(datetime)").get()
        description = "\n".join(response.css(".field--name-body p::text").getall())
        terms_html = response.css("#report-terms .field")
        terms = {}
        for term in terms_html:
            key = term.css(".field__label::text").get()
            value = term.css(".field__item a::text").getall()
            terms[key] = value
        file_urls = response.css(".field--type-file a::attr(href)").getall()
        rel_paths = [
            str(self.get_file_path(url, report_id).relative_to(self.output)) for url in file_urls
        ]
        meta = ReportMeta(
            **{
                "report_id": report_id,
                "title": title,
                "timestamp": timestamp,
                "description": description,
                "terms": terms,
                "files": rel_paths,
                "file_urls": file_urls,
                "url": response.url,
            }
        )
        yield meta
        for file_url in file_urls:
            yield response.follow(
                file_url, callback=self.parse_report_file, cb_kwargs={"report_id": report_id}
            )

    def parse_report_file(self, response: Response, report_id: str) -> None:
        out_path = self.get_file_path(response.url, report_id)
        with open(out_path, "wb") as f:
            f.write(response.body)

    def parse_archive_urls(self, urls: list[str]) -> Generator[Request, None, None]:
        self.logger.info("Got archive urls: %s", urls)
        urls = [url for url in urls if not (self.output / "archive" / url.split("/")[-1]).exists()]
        for url in urls:
            yield scrapy.Request(url, callback=self.parse_archive_pdf)

    def parse_archive_pdf(self, response: Response) -> None:
        report_path = self.output / "archive" / response.url.split("/")[-1]
        with open(report_path, "wb") as f:
            f.write(response.body)

    def get_file_path(self, url: str, report_id: str) -> Path:
        report_path = self.output / "files" / report_id
        report_path.mkdir(exist_ok=True)
        return report_path / url.split("/")[-1]



import random
from pathlib import Path

import bagit
import pytest

from sciop_scraping.validation import pick_bagit_manifest, validate_bagit_manifest


@pytest.mark.parametrize("algo", ["sha256", "sha512", "md5"])
def test_valid_bagit(tmp_bagit: bagit.Bag, algo: str):
    """a valid bagit archive should be valid..."""
    errors = validate_bagit_manifest(tmp_bagit.path, algo)
    assert len(errors) == 0


@pytest.mark.parametrize("algo", ["sha256", "sha512", "md5"])
def test_invalid_bagit(tmp_bagit: bagit.Bag, algo: str):
    """an invalid bagit archive should not be valid..."""
    path = Path(tmp_bagit.path)
    with open(path / "data" / "regular.txt", "wb") as f:
        f.write(random.randbytes(32 * (2**10)))

    errors = validate_bagit_manifest(tmp_bagit.path, algo)
    assert len(errors) == 1
    assert errors[0].path == Path("data/regular.txt")
    assert errors[0].type_ == "incorrect"


def test_pick_manifest(tmp_path: Path) -> None:
    paths = [
        "manifest-md5.txt",
        "manifest-sha1.txt",
        "manifest-sha2.txt",
        "manifest-sha256.txt",
        "manifest-sha512.txt",
        "manifest-fakealgo.txt",
        "otherfile.txt",
    ]
    priority = ["sha1", "sha256", "sha512", "md5"]
    for path in paths:
        (tmp_path / path).write_text("sup")

    for _ in range(len(priority)):
        last = priority.pop(0)
        priority.append(last)
        manifest = pick_bagit_manifest(tmp_path, priority)
        assert manifest[0] == priority[0]
        assert manifest[0] in str(manifest[1])
        assert manifest[1].exists()


def test_reject_fake_algos(tmp_path: Path) -> None:
    """
    We can specify a custom algorithm priority, but if python doesn't have it in hashlib,
    we still error
    """
    algo = "fakealgo"
    (tmp_path / f"manifest-{algo}.txt").write_text("sup")
    with pytest.raises(FileNotFoundError):
        pick_bagit_manifest(tmp_path, [algo])


def test_detect_unranked_algos(tmp_path: Path) -> None:
    """
    we fall back to algos if prioritized ones aren't found, but ones in hashlib are
    """
    algo = "fakealgo"
    (tmp_path / "manifest-sha512.txt").write_text("sup")
    manifest = pick_bagit_manifest(tmp_path, [algo])
    assert manifest[0] == "sha512"



from abc import ABC, abstractmethod
from collections.abc import Generator
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Self
from urllib.parse import parse_qs, urlencode, urljoin, urlparse, urlunparse

import scrapy


@dataclass
class S3V2ListItem:
    key: str
    last_modified: datetime
    etag: str
    size: int
    storage_class: str
    checksum_algorithm: str | None = None
    checksum_type: str | None = None
    owner: dict | None = None
    restore_status: dict | None = None

    @classmethod
    def from_xml(cls, item: scrapy.Selector) -> Self:
        return S3V2ListItem(
            key=item.xpath("s3:Key/text()").get(),
            last_modified=datetime.fromisoformat(item.xpath("s3:LastModified/text()").get()),
            etag=item.xpath("s3:ETag/text()").get().lstrip('"').rstrip('"'),
            size=int(item.xpath("s3:Size/text()").get()),
            storage_class=item.xpath("s3:StorageClass/text()").get(),
            checksum_algorithm=item.xpath("s3:ChecksumAlgorithm/text()").get(),
            checksum_type=item.xpath("s3:ChecksumType/text()").get(),
            owner=item.xpath("s3:Owner/text()").get(),
            restore_status=item.xpath("s3:RestoreStatus/text()").get(),
        )


class S3Mixin(ABC):
    def iter_s3_items(
        self, bucket_url: str, prefix: str | None = None
    ) -> Generator[scrapy.Request]:
        """
        Iterate through pages in an AWS bucket (given as the URL of the bucket)

        Yields requests to two parsing callbacks:
        - `parse_s3_page`: defined here, receive the page and yield a request for the next page
        - `parse_s3_items`: Must be defined in a subclass -
          receives the actual items to be used for further processing
        """
        query = {
            "list-type": 2,
            "encoding-type": "url",
        }
        if prefix is not None:
            query["prefix"] = prefix
        url = urljoin(bucket_url, "?" + urlencode(query))
        yield scrapy.Request(url, callback=self.parse_s3_page)

    def parse_s3_page(self, response: scrapy.http.XmlResponse) -> Generator[scrapy.Request]:
        # yield the request for the next page
        parsed = urlparse(response.request.url)
        query = parse_qs(parsed.query)
        selector = scrapy.Selector(response, type="xml")
        selector.register_namespace("s3", "http://s3.amazonaws.com/doc/2006-03-01/")
        next_token = selector.xpath("//s3:NextContinuationToken/text()").get()
        if next_token:
            # this is seriously so annoying to have to do
            query = {k: v[0] for k, v in query.items()}
            query["continuation-token"] = next_token
            encoded_query = urlencode(query)
            unparsed = urlunparse(
                (
                    parsed.scheme,
                    parsed.netloc,
                    parsed.path,
                    parsed.params,
                    encoded_query,
                    parsed.fragment,
                )
            )
            yield scrapy.Request(unparsed, priority=1, callback=self.parse_s3_page)

        # call the overridden callback with the items
        items = selector.xpath("//s3:Contents")
        items_cast = [S3V2ListItem.from_xml(item) for item in items]
        yield from self.parse_s3_items(response, items_cast)

    @abstractmethod
    def parse_s3_items(self, response: scrapy.http.Response, items: list[S3V2ListItem]) -> Any: ...



from sciop_cli.exceptions import SciopException


class QuestError(SciopException):
    """Generic parent class for all quest-related errors"""


class QuestUploadError(SciopException):
    """Parent exception for quest upload errors"""


class QuestValidationError(QuestError, ValueError):
    """Error emitted when a quest is invalid for the attempted action"""


class QuestInvalidUploadError(QuestUploadError, QuestValidationError):
    """A quest could not be uploaded because it was invalid or incomplete!"""



from pathlib import Path

from rich.console import Console

from sciop_scraping.quests.base import QuestLog, QuestStatus, ValidationError


def test_status_table(tmp_path: Path):
    """
    Literally just check that the table prints, not trying to do ser/des for the table atm.
    """
    log = QuestLog(
        subquests=[
            QuestStatus(
                quest="sup",
                dataset_slug="sup",
                subquest="hey",
                path=tmp_path / "hey",
                validation_errors=[
                    ValidationError(
                        path=tmp_path / "hey" / "howdy.exe", type="uhh", msg="it was bad"
                    )
                ],
                status="scraping",
                result=None,
            ),
            QuestStatus(
                quest="sup",
                dataset_slug="sup",
                subquest="hello",
                path=tmp_path / "hello",
                status="complete",
                result="success",
            ),
            QuestStatus(
                quest="goodbye",
                dataset_slug="bye",
                subquest="bye",
                path=tmp_path / "bye",
                status="uploaded",
                result="success",
            ),
        ]
    )
    _ = log.to_tables()
    console = Console()
    console.print(log)



from pathlib import Path

import click
from scrapy.crawler import CrawlerProcess

from sciop_scraping.quests.poes.spider import PoesSpider


@click.command("noaa-poes")
@click.option(
    "-o",
    "--output",
    help="Output directory to save files in. "
    "If None, $PWD/data/noaa-poes. "
    "Data will be saved in a chronicling-america subdirectory, "
    "and the crawl state will be saved in crawl_state.",
    default=None,
    type=click.Path(),
)
def noaa_poes(output: str | None = None) -> None:
    output = Path.cwd() / "data" / "noaa-poes" if not output else Path(output)
    output.mkdir(exist_ok=True, parents=True)
    process = CrawlerProcess()
    process.crawl(PoesSpider, output=output)
    process.start()



import hashlib
import multiprocessing as mp
import random
import re
from collections.abc import AsyncIterator, Generator
from io import BytesIO
from pathlib import Path
from typing import Any, cast
from urllib.parse import urljoin

import scrapy
from scrapy.http import HtmlResponse, Response, TextResponse
from scrapy.settings import BaseSettings
from scrapy.spidermiddlewares.httperror import HttpError
from tqdm import tqdm
from twisted.python.failure import Failure

from sciop_scraping.const import USER_AGENTS
from sciop_scraping.validation import BAGIT_ALGO_PRIORITY, BAGIT_MANIFEST

BASE_URL = "https://chroniclingamerica.loc.gov/data/batches/"
# standard files that we will look for in all cases
ALL_BAGIT_FILES = ["bag-info.txt", "bagit.txt"]
# data manifest files - we expect at least one
ALL_MANIFEST_FILES = [f"manifest-{hash_type}.txt" for hash_type in BAGIT_ALGO_PRIORITY]
# tagmanifest files - optional, but if present, they may contain other metadata filenames
ALL_TAGMANIFEST_FILES = [f"tag{file}" for file in ALL_MANIFEST_FILES]
# initial list to try
ALL_META_FILES = ALL_BAGIT_FILES + ALL_MANIFEST_FILES + ALL_TAGMANIFEST_FILES


class ChroniclingAmericaSpider(scrapy.Spider):
    """
    Crawl chronicling america data from the Library of Congress,
    either by batch or as a whole.

    Excludes `.tif` and `.tiff` files, since they are redundant with the .jp2 files.
    """

    name = "chronicling-america"
    allowed_domains = ["chroniclingamerica.loc.gov", "tile.loc.gov"]

    USER_AGENT_OVERRIDE: str | None = None
    JOBDIR_OVERRIDE: str | None = None
    custom_settings = {
        "RETRY_ENABLED": True,
        "RETRY_TIMES": 100,
        "RETRY_HTTP_CODES": [500, 502, 503, 504, 522, 524, 408],
        "CONCURRENT_REQUESTS": int(round(mp.cpu_count() * 1.5)),
        "CONCURRENT_REQUESTS_PER_DOMAIN": int(round(mp.cpu_count() * 1.5)),
    }

    def __init__(
        self,
        batch: str | None = None,
        output: Path | None = None,
        cf_cookie: str | None = None,
        download_timeout: int = 20,
        resume: bool = False,
        *args: Any,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if batch:
            self.batch = batch.replace("-", "_")
        else:
            self.batch = batch
        if output is None:
            self.output = Path.cwd() / "data" / "chronicling-america"
        else:
            self.output = Path(output).resolve()
        self.output.mkdir(exist_ok=True, parents=True)
        self.cf_cookie = {"cf_clearance": cf_cookie} if cf_cookie else None
        self.crawl_root = self.crawl_root(self.batch)
        if not self.batch:
            self.start_urls = [BASE_URL]
        else:
            self.start_urls = [urljoin(self.crawl_root, f) for f in ALL_META_FILES]
        self.download_timeout = download_timeout
        self.resume = resume
        # logging.getLogger('scrapy').propagate = False

    async def start(self) -> AsyncIterator[Any]:

        for url in self.start_urls:
            # tagmanifest is an optional manifest for other meta files.
            # we parse up to one in case there are any unexpected meta files to add to the url list.
            # we will download all tagmanifest files
            if url.split("/")[-1] in ALL_TAGMANIFEST_FILES:
                yield scrapy.Request(
                    self.tile_url(url),
                    callback=self.parse_tagmanifest,
                    dont_filter=True,
                    cookies=self.cf_cookie,
                    errback=self.errback,
                    cb_kwargs={"batch_url": self.crawl_root, "original_url": url},
                )

            # manifest is the main manifest for the data in the batch.
            # there should be one, there could be more (different hashing algorithms).
            # we parse the first one and download any others.
            elif url.split("/")[-1] in ALL_MANIFEST_FILES:
                yield scrapy.Request(
                    self.tile_url(url),
                    callback=self.parse_manifest,
                    dont_filter=True,
                    cookies=self.cf_cookie,
                    errback=self.errback,
                    cb_kwargs={"batch_url": self.crawl_root, "original_url": url},
                )

            # if there is no tagmanifest, we want to at least get certain predefined files anyway
            elif url.split("/")[-1] in ALL_BAGIT_FILES:
                yield scrapy.Request(
                    self.tile_url(url),
                    callback=self.parse_file,
                    dont_filter=True,
                    cookies=self.cf_cookie,
                    errback=self.errback,
                    cb_kwargs={"original_url": url},
                )

            # get any other data files
            # (including additional manifest/tagmanifest files after we have processed one)
            else:
                yield scrapy.Request(
                    url, dont_filter=True, cookies=self.cf_cookie, errback=self.errback
                )

    @classmethod
    def update_settings(cls, settings: BaseSettings) -> None:
        super().update_settings(settings)
        if cls.USER_AGENT_OVERRIDE:
            settings.set("USER_AGENT", cls.USER_AGENT_OVERRIDE, priority="spider")
        else:
            settings.set("USER_AGENT", random.choice(USER_AGENTS), priority="spider")
        if cls.JOBDIR_OVERRIDE:
            settings.set("JOBDIR", str(cls.JOBDIR_OVERRIDE))
        settings.set("ROBOTSTXT_OBEY", False, priority="spider")
        # there doesn't appear to be a rate limit on tile.loc.gov,
        # and since we can jump straight there when we have a batch name,
        # and i'm not sure how to dynamically modify this, leave this off by default.
        # settings.set("DOWNLOAD_DELAY", 0.5, priority="spider")

    def url_to_path(self, url: str) -> Path:
        """Get the output path for a given URL."""
        out_name = re.sub(BASE_URL, "", url)
        out_path = self.output / out_name
        out_path.parent.mkdir(exist_ok=True, parents=True)

        return out_path

    @staticmethod
    def crawl_root(batch: str | None = None) -> str:
        if not batch:
            return BASE_URL
        else:
            return BASE_URL + batch + "/"

    @staticmethod
    def tile_url(url: str) -> str:
        """Convert a chroniclingamerica.loc.gov url to a tile.loc.gov url"""
        if "tile.loc.gov" in url:
            return url
        pattern = re.compile(
            r"^https://chroniclingamerica\.loc\.gov/data/batches/(?P<batch>\w+)/(?P<path>.*)"
        )
        match = pattern.match(url)
        if not match:
            raise ValueError(f"Could not convert url to tile url: {url}")
        val = match.groupdict()
        prefix = val["batch"].split("_")[0]
        return f"https://tile.loc.gov/storage-services/service/ndnp/{prefix}/batch_{val['batch']}/{val['path']}"

    @staticmethod
    def webseed_urls(batch: str) -> list[str]:
        """
        Return a pair of URLs that can be used as webseed urls in created torrents.

        like:
        - https://chroniclingamerica.loc.gov/data/batches/
        - https://tile.loc.gov/storage-services/service/ndnp/{batch_prefix}/batch_
        """
        tile_url = ChroniclingAmericaSpider.tile_url(ChroniclingAmericaSpider.crawl_root(batch))
        # strip the part so that when we append the batch subdirectory we get the full url
        parts = re.split(r"(?<=batch_)", tile_url)
        return [BASE_URL, parts[0]]

    def parse(self, response: HtmlResponse) -> Generator[scrapy.Request, None, None]:
        links = response.css("a::attr(href)").getall()

        # TODO tagmanifest parsing for this scenario

        # if we are on a page that has a `manifest-*.txt` in it,
        # shortcut and just use that to enumerate the files
        manifest_links = [link for link in links if BAGIT_MANIFEST.fullmatch(link.split("/")[-1])]
        if manifest_links:
            for link in manifest_links:
                link = urljoin(response.url, link)
                yield response.follow(
                    link,
                    callback=self.parse_manifest,
                    errback=self.errback,
                    cookies=self.cf_cookie,
                    cb_kwargs={"batch_url": response.url, "original_url": link},
                )
        else:
            links = [urljoin(response.url, link) for link in links]
            # filter to only those underneath the crawl root
            links = [link for link in links if self.crawl_root in link]

            for link in links:
                if link.endswith("/"):
                    yield response.follow(
                        link, callback=self.parse, errback=self.errback, cookies=self.cf_cookie
                    )
                else:
                    if not self.url_to_path(link).exists():
                        tile_link = self.tile_url(link)
                        yield response.follow(
                            tile_link,
                            callback=self.parse_file,
                            errback=self.errback,
                            cb_kwargs={"original_url": link},
                            cookies=self.cf_cookie,
                        )
                    else:
                        self.logger.info("Skipping %s, already downloaded", link)

    def parse_tagmanifest(self, response: TextResponse, batch_url: str, original_url: str) -> None:
        """
        parse the first tagmanifest file (any others will be downloaded without parsing).
        we only need to enqueue any files listed in tagmanifest which are not in
        MANIFEST_FILES or BAGIT_FILES.
        """
        self.parse_file(response, original_url)

        lines = response.text.split("\n")
        path_hashes = [re.split(r"\s+", line) for line in lines if line]
        urls = [(p[0], urljoin(batch_url, p[1])) for p in path_hashes if p[1] not in ALL_META_FILES]

        self.logger.info(
            f"Downloading {len(urls)} additional unexpected files from tagmanifest {response.url}"
        )
        for url in urls:
            tile_url = self.tile_url(url[1])
            yield response.follow(
                tile_url,
                callback=self.parse_file,
                errback=self.errback,
                cb_kwargs={"original_url": url[1]},
                cookies=self.cf_cookie,
            )

    def parse_manifest(self, response: TextResponse, batch_url: str, original_url: str) -> None:
        # save the manifest
        manifest_file = original_url.split("/")[-1]
        hash_type = BAGIT_MANIFEST.match(manifest_file).groupdict()["hash_type"]
        self.parse_file(response, original_url=urljoin(batch_url, manifest_file))

        lines = response.text.split("\n")
        path_hashes = [re.split(r"\s+", line) for line in lines if line]
        # exclude .tif and .tiff files
        paths = [
            p for p in path_hashes if not p[-1].endswith(".tif") and not p[-1].endswith(".tiff")
        ]
        urls = [(p[0], urljoin(batch_url, p[1])) for p in paths]

        # separate completed, partial, and not started files
        # yeah this is a little inefficient but we only need to do it once
        existing = [url for url in urls if self.url_to_path(url[1]).exists()]
        not_existing = [url for url in urls if not self.url_to_path(url[1]).exists()]

        self.logger.info(f"Downloading {len(not_existing)} new files")

        for url in not_existing:
            tile_url = self.tile_url(url[1])
            yield response.follow(
                tile_url,
                callback=self.parse_file,
                errback=self.errback,
                cb_kwargs={"original_url": url[1]},
                cookies=self.cf_cookie,
            )

        # check existing files for completeness, retry partial files with byte offset ranges
        if existing and self.resume:
            self.logger.info(f"Checking {len(existing)} existing files for completion")
            yield from self._resume(existing, response, hash_type)
        else:
            return

    def _resume(
        self, existing: list[tuple[str, str]], response: TextResponse, hash_type: str
    ) -> None:
        partial = []

        for digest_url in tqdm(existing):
            expected_digest, url = digest_url
            path = self.url_to_path(url)
            with open(path, "rb") as f:
                digest = hashlib.file_digest(f, hash_type).hexdigest()
            if digest == expected_digest:
                continue
            partial.append(url)
            existing_size = path.stat().st_size
            yield response.follow(
                self.tile_url(url),
                callback=self.parse_partial_file,
                errback=self.errback,
                cb_kwargs={"original_url": url},
                cookies=self.cf_cookie,
                headers={"Range": f"bytes={existing_size}-"},
            )

        if partial:
            self.logger.info(f"Retrying {len(partial)} partial downloads")
        self.logger.info(f"Not downloading {len(existing) - len(partial)} completed files")

    def parse_file(self, response: Response, original_url: str) -> None:
        """
        Save a file underneath ``self.output``, removing the base url
        """
        out_path = self.url_to_path(original_url)
        self.logger.debug("Saving %s to %s", response.url, out_path)
        with open(out_path, "wb") as f:
            f.write(response.body)

    def parse_partial_file(self, response: Response, original_url: str) -> None:
        """
        Append to an already-existing partial file
        """
        out_path = self.url_to_path(original_url)

        # if partial range returned, append, otherwise write new
        mode = "ab" if response.status == 206 else "wb"
        if mode == "ab":
            self.logger.debug("Appending partial file %s to %s", response.url, out_path)
        else:
            self.logger.debug(
                "HTTP range ignored by server, saving %s to %s", response.url, out_path
            )
        with open(out_path, mode) as f:
            f: BytesIO
            f.write(response.body)

    def errback(self, failure: Failure) -> None:
        self.logger.error(f"A spider error occurred: {failure}")
        if failure.check(HttpError) and failure.value.response.status == 429:
            raise scrapy.exceptions.CloseSpider(
                "429 received - stopping scrape. We can't wait these out. \n"
                "You should \n"
                "- Open one of the recent pages in a browser,\n"
                "- Pass the cloudflare check\n"
                "- Open your developer tools (often right click + inspect element)\n"
                "- Open the networking tab to watch network requests\n"
                "- Reload the page\n"
                "- Click on the request made to the page you're on to see the request headers\n"
                "- Copy your user agent and the part of the cookie after `cf_clearance=` "
                "and pass them to the -u and -c cli options, respectively.\n"
                "If you are crawling *all* data, rather than a single batch, "
                "you will likely need to set DOWNLOAD_DELAY=1 until the manifests are scraped."
            )
        elif failure.check(HttpError) and failure.value.response.status == 416:
            # requested an impossible range, e.g. from when we have an incorrect hash
            # and try to resume. Resolves ambiguity between incomplete vs. incorrect hash mismatches
            failure.value.response = cast(Response, failure.value.response)
            if "original_url" in failure.request.cb_kwargs:
                original_url = failure.request.cb_kwargs["original_url"]
                path = self.url_to_path(original_url)

                self.logger.warning(
                    "Removing file with unsatisfiable request range %s from %s and retrying",
                    path,
                    original_url,
                )
                path.unlink(missing_ok=True)
                yield scrapy.Request(
                    self.tile_url(original_url),
                    callback=self.parse_file,
                    errback=self.errback,
                    cb_kwargs={"original_url": original_url},
                    cookies=self.cf_cookie,
                    dont_filter=True,
                )



import multiprocessing as mp
import posixpath
import re
from collections.abc import AsyncIterator
from itertools import product
from pathlib import Path
from typing import Any

import scrapy
from scrapy.http import HtmlResponse, Response

URL_ROOT = "https://www.ncei.noaa.gov/data/poes-metop-space-environment-monitor/access/"
L1A_URL = "https://www.ncei.noaa.gov/data/poes-metop-space-environment-monitor/access/l1a/v01r00/"
L1B_URL = "https://www.ncei.noaa.gov/data/poes-metop-space-environment-monitor/access/l1b/v01r00/"
DECOMISSIONED = ("noaa15", "noaa19")
YEARS = [year for year in range(2012, 2026)]
IGNORE_LINKS = ("Parent Directory", "Name", "Last modified", "Size")


class PoesSpider(scrapy.Spider):
    """
    https://www.ncei.noaa.gov/products/poes-metop-space-environment-monitor
    """

    name = "noaa-poes"
    custom_settings = {
        "RETRY_ENABLED": True,
        "RETRY_TIMES": 10,
        "RETRY_HTTP_CODES": [500, 502, 503, 504, 522, 524, 408],
        "CONCURRENT_REQUESTS": int(round(mp.cpu_count() * 1.5)),
        "CONCURRENT_REQUESTS_PER_DOMAIN": int(round(mp.cpu_count() * 1.5)),
    }

    def __init__(self, output: Path):
        self.output = output

    def url_to_path(self, file_url: str) -> Path:
        relative = re.sub(URL_ROOT, "", file_url)
        return self.output / relative

    async def start(self) -> AsyncIterator[Any]:
        for group, year, sat in product((L1A_URL, L1B_URL), YEARS, DECOMISSIONED):
            full_url = posixpath.join(group, str(year), sat) + "/"
            yield scrapy.Request(
                full_url,
                callback=self.parse_files,
                dont_filter=True,
            )

    async def parse_files(self, response: HtmlResponse) -> None:
        links = response.css("td > a")
        for link in links:
            if link.css("a::text").get() in IGNORE_LINKS:
                continue
            file_url = response.urljoin(link.attrib["href"])
            file_path = self.url_to_path(file_url)
            if file_path.exists():
                continue
            file_path.parent.mkdir(exist_ok=True, parents=True)
            yield response.follow(file_url, callback=self.parse_file)

    def parse_file(self, response: Response) -> None:
        """
        Save a file underneath ``self.output``, removing the base url
        """

        out_path = self.url_to_path(response.url)
        self.logger.debug("Saving %s to %s", response.url, out_path)
        with open(out_path, "wb") as f:
            f.write(response.body)



import re
from pathlib import Path
from typing import Any, Literal

import click
from rich import print as rprint
from sciop_cli.api import claim_next
from sciop_cli.config import get_config

from sciop_scraping.quests import QuestLog
from sciop_scraping.quests.base import upload_subquest
from sciop_scraping.quests.smithsonian.const import BUCKET_URL, DATASETS
from sciop_scraping.quests.smithsonian.quest import SmithsonianQuest


@click.command("smithsonian")
@click.option("-o", "--output", type=click.Path(dir_okay=True, file_okay=False), default=None)
@click.option("-d", "--dataset", type=click.Choice(DATASETS), required=False)
@click.option("-f", "--format", type=click.Choice(("jpg", "tif")), required=False)
@click.option(
    "--next",
    help="Claim, download, and upload the next available dataset part from the quest. "
    "Must have already logged in with `sciop-cli login` or `sciop-scrape api login`. "
    "Mutually exclusive with --batch and --auto.",
    is_flag=True,
    default=False,
)
@click.option(
    "--upload/--no-upload",
    help="Create a torrent and upload it to the configured sciop instance.",
    default=True,
    is_flag=True,
    show_default=True,
)
def smithsonian(
    output: Path = None,
    dataset: str | None = None,
    format: Literal["jpg", "tif"] | None = None,
    next: bool = False,
    upload: bool = True,
) -> None:
    """
    Batches from the smithsonian public s3 bucket
    """
    output = Path.cwd() / "data" if output is None else Path(output)
    torrent_dir = output / "torrents"

    if not next and (not dataset or not format):
        raise ValueError("If not using --next, must specify dataset and format explicitly")
    elif next:
        dataset_slug, format = _next_batch(output, upload)
        dataset = re.sub(r"^si-", "", dataset_slug)
        if dataset is None:
            return
    else:
        dataset_slug = "si-" + dataset

    quest = SmithsonianQuest(
        output=output,
        dataset_slug=dataset_slug,
        subquest=f"{dataset}-{format}",
        format=format,
        part_slugs=[format],
    )
    quest.meta_path.parent.mkdir(parents=True, exist_ok=True)
    res = quest.run(
        process_kwargs={
            "settings": {"FEEDS": {quest.meta_path: {"format": "jsonlines", "overwrite": True}}}
        }
    )
    if upload:
        torrent_dir.mkdir(exist_ok=True)
        upload, torrent_path = upload_subquest(
            status=res,
            torrent_dir=torrent_dir,
            torrent_path=torrent_dir / f"si-{dataset}-{format}.torrent",
            progress=True,
            torrent_kwargs={"webseeds": [BUCKET_URL + "/media/"]},
        )
        click.echo(f"Uploaded {res.quest} - {res.subquest}")
        rprint(upload)

        _upload_to_client(
            torrent_path=torrent_path, data_path=res.path.parent, tags=[res.quest, "sciop-scraper"]
        )
        res.status = "uploaded"
        quest_log = QuestLog.from_json(output / "quest-log.json")
        quest_log.update(res)
        quest_log.to_json(path=output / "quest-log.json")
    rprint(res)


def _next_batch(output: Path, upload: bool) -> tuple[str, str] | tuple[None, None]:
    log = QuestLog.from_json(output / "quest-log.json")
    if upload:
        existing = [
            subq
            for subq in log.subquests
            if subq.quest == "smithsonian"
            and (subq.status not in ("uploaded",) or subq.result != "success")
        ]
    else:
        existing = [
            subq
            for subq in log.subquests
            if subq.quest == "smithsonian"
            and (subq.status not in ("complete", "uploaded") or subq.result != "success")
        ]

    # filter impossible or disabled items
    existing = [e for e in existing if e.status != "disabled" and e.result != "impossible"]

    if len(existing) > 0:
        subquest = existing[0].subquest
        click.echo(f"Resuming previous incomplete subquest {subquest}")
        return existing[0].dataset_slug, existing[0].part_slugs[0]
    else:
        claim = None
        for ds in DATASETS:
            claim = claim_next(f"si-{ds}")
            if claim:
                click.echo(f"Claimed next subquest: {claim['dataset']} - {claim['dataset_part']}")
                return claim["dataset"], claim["dataset_part"]
        click.echo("No unclaimed dataset parts remaining! quest complete!")
        return None, None


def _upload_to_client(torrent_path: Path, data_path: Path, **kwargs: Any) -> None:
    """Upload to a locally configured bittorrent client, if any"""
    cfg = get_config()
    clients = [c for c in cfg.clients if c.host in ("localhost", "127.0.0.1")]
    if not clients:
        click.echo(
            "No local bittorrent clients configured, so not seeding yet. "
            f"Add the torrent {torrent_path} to your client manually. "
            "use `sciop-cli client login` to add a client!"
        )
        return

    for client in clients:
        try:
            adapter = client.make_adapter()
            adapter.add_torrent_file(torrent_path=torrent_path, data_path=data_path, **kwargs)
            click.echo(
                f"Uploaded {torrent_path} to {client.client} client at {client.host}:{client.port}"
            )
        except Exception as e:
            click.echo(
                f"Error trying to upload {torrent_path} to {client.client} at "
                f"{client.host}:{client.port}, "
                f"Got exception: \n{e}",
                err=True,
            )



"""
Helper methods for validation that might be used in multiple quests
"""

import hashlib
import re
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, cast

from tqdm import tqdm

if TYPE_CHECKING:
    from sciop_scraping.quests.base import ValidationError

BAGIT_MANIFEST = re.compile(r"^manifest-(?P<hash_type>\w+).txt$")
BAGIT_ALGO_PRIORITY = ("sha512", "sha256", "sha1", "md5")


class _BagitManifest(NamedTuple):
    hash_type: str
    path: Path


def pick_bagit_manifest(
    path: Path, algo_priority: tuple[str, ...] | list[str] = BAGIT_ALGO_PRIORITY
) -> _BagitManifest:
    """
    Given a bagit directory and a ranking of hash algorithms,
    pick the highest ranked algo, if any from the priority list are present.
    Otherwise, return any that are present.

    .. note:: Cursed implementation

        the implementation is in honor of a prior code golf implementation,
        and while it works, it should not serve as an indication of the
        attitude of the authors towards the mental health of its readers.

    raises:
        FileNotFoundError if no bagit manifests are present
    """
    manifests: list[_BagitManifest] = sorted(
        [
            _BagitManifest(hash_type=cast(str, match.group("hash_type")), path=p)
            for p in Path(path).iterdir()
            if p.is_file() and (match := BAGIT_MANIFEST.fullmatch(p.name))
        ],
        key=lambda manifest: (
            (in_ranking := manifest[0] in algo_priority),
            algo_priority.index(manifest[0]) if in_ranking else 0,
        ),
    )
    manifests = [m for m in manifests if m[0] in hashlib.algorithms_available]
    if not manifests:
        raise FileNotFoundError(f"No bagit manifests with supported hash algo found in {path}")
    return manifests[0]


def validate_bagit_manifest(
    path: Path, hash_type: str = "md5", remove: bool = False, hash_when_missing: bool = True
) -> list["ValidationError"]:
    """
    Given the base directory of a bagit directory that contains `manifest-{hash_type}.txt`
    and `data/` check the files against the manifest,
    returning ValidationErrors for missing or incorrect files.

    Args:
        path (Path): bagit directory
        hash_type (str): string name of hash algo,
            should match the file name abbreviation and be available by that name in hashlib
        remove (bool): If ``True``, remove invalid files (default False)
        hash_when_missing (bool): If ``False``, when files are missing, skip hash checking.
            If ``True``, hash files even if some are missing
    """
    from sciop_scraping.quests.base import ValidationError

    errors = []
    path = Path(path)
    manifest_path = path / f"manifest-{hash_type}.txt"

    if not manifest_path.exists():
        errors.append(
            ValidationError(
                type="manifest",
                path=manifest_path,
                msg="No manifest file found at expected location!",
            )
        )
        return errors

    with open(manifest_path) as f:
        manifest = f.read()

    lines = manifest.splitlines()
    # split into (hash, path) pairs
    items = [re.split(r"\s+", line.strip(), maxsplit=1) for line in lines]

    # first check for missing files, we know we're invalid if we have missing files
    # and can quit early
    for item in tqdm(items, desc="Checking for missing files"):
        expected_hash, sub_path = item
        abs_path = path / sub_path
        if not abs_path.exists():
            errors.append(
                ValidationError(type="missing", path=Path(sub_path), msg="File not found")
            )

    if not hash_when_missing and len(errors) > 0:
        return errors

    for item in tqdm(items, desc="Validating file hashes"):
        expected_hash, sub_path = item
        abs_path = path / sub_path
        with open(abs_path, "rb") as f:
            file_hash = hashlib.file_digest(f, hash_type).hexdigest()

        if file_hash != expected_hash:
            errors.append(
                ValidationError(type="incorrect", path=Path(sub_path), msg="Hash mismatch")
            )
            if remove:
                abs_path.unlink()
    return errors



import multiprocessing as mp
import re
from collections.abc import AsyncGenerator
from pathlib import Path
from typing import Any, Literal
from urllib.parse import urljoin

import scrapy

from sciop_scraping.mixins.s3 import S3Mixin, S3V2ListItem
from sciop_scraping.quests.smithsonian.const import BUCKET_URL


class SmithsonianSpider(scrapy.Spider, S3Mixin):
    name = "smithsonian"

    custom_settings = {
        "RETRY_ENABLED": True,
        "RETRY_TIMES": 10,
        "RETRY_HTTP_CODES": [422, 500, 502, 503, 504, 522, 524, 408],
        "CONCURRENT_REQUESTS": int(round(mp.cpu_count() * 1.5)),
        "CONCURRENT_REQUESTS_PER_DOMAIN": int(round(mp.cpu_count() * 1.5)),
        "DOWNLOAD_WARNSIZE": 1 * (2**30),  # 1 GiB
        "DOWNLOAD_TIMEOUT": 60 * 15,
        "DOWNLOAD_MAXSIZE": 0,
    }

    def __init__(self, output: Path, dataset: str, format: Literal["jpg", "tif"]):
        super().__init__()
        self.output = Path(output)
        self.dataset = dataset
        self.format = format

    async def start(self) -> AsyncGenerator[scrapy.Request]:
        yield next(self.iter_s3_items(BUCKET_URL, "media/" + self.dataset))

    def parse_s3_items(self, response: scrapy.http.Response, items: list[S3V2ListItem]) -> Any:
        for item in items:
            path = self.path_from_key(item.key)

            if self.format == "jpg" and path.suffix not in (".jpg", ".jpeg"):  # noqa: SIM114
                continue
            elif self.format == "tif" and path.suffix not in (".tif", ".tiff"):
                continue

            yield item

            if path.exists():
                continue

            yield scrapy.Request(
                urljoin(BUCKET_URL, item.key), callback=self.parse_file, cb_kwargs={"item": item}
            )

    def parse_file(self, response: scrapy.http.Response, item: S3V2ListItem) -> None:
        path = self.path_from_key(item.key)
        with open(path, "wb") as f:
            f.write(response.body)

    def path_from_key(self, key: str) -> Path:
        key = re.sub(r"^media/", "", key)
        path = self.output / "smithsonian" / f"{self.dataset}-{self.format}" / key
        path.parent.mkdir(parents=True, exist_ok=True)
        return path
